import os
import re
import tempfile
import datetime

from Ganga.GPIDev.Credentials import credential_store

from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger

from Ganga.Utility.GridShell import getShell

from Ganga.Lib.LCG.GridftpSandboxCache import GridftpFileIndex, GridftpSandboxCache

from Ganga.Lib.LCG.Utility import get_uuid
from Ganga.Lib.Root import randomString

# global variables
logger = getLogger()

config = getConfig('LCG')


def __set_submit_option__():

    submit_option = ''

    if config['Config']:
        submit_option += ' --config %s' % config['Config']
    elif config['GLITE_ALLOWED_WMS_LIST']:
        wms_conf_path = os.path.join(os.environ['GLITE_WMS_LOCATION'], 'etc', config['VirtualOrganisation'], 'glite_wmsui.conf')
        temp_wms_conf = tempfile.NamedTemporaryFile(suffix='.conf', delete=False)

        with open(wms_conf_path, "r") as this_file:
            orig_text = this_file.read()

        # find the last bracket and add in the new text
        pos = orig_text.rfind("]")
        wms_text = "\nWMProxyEndpoints  =  {" + \
                   ",".join("\"%s\"" % wms for wms in config['GLITE_ALLOWED_WMS_LIST']) + \
                   "};\n]\n"
        new_text = orig_text[:pos] + wms_text

        # write the new config file
        with open(temp_wms_conf, "w") as this_file:
            this_file.write(new_text)

        submit_option += ' --config %s' % temp_wms_conf.name

    submit_option = ' %s ' % submit_option

    return submit_option


def __resolve_gridcmd_log_path__(regxp_logfname, cmd_output):
    match_log = re.search(regxp_logfname, cmd_output)

    logfile = None
    if match_log:
        logfile = match_log.group(1)
    return logfile


def __clean_gridcmd_log__(regxp_logfname, cmd_output):

    logfile = __resolve_gridcmd_log_path__(regxp_logfname, cmd_output)

    if logfile and os.path.exists(logfile):
        os.remove(logfile)

    return True


def __print_gridcmd_log__(regxp_logfname, cmd_output):

    logfile = __resolve_gridcmd_log_path__(regxp_logfname, cmd_output)

    if logfile:
        for l in open(logfile, 'r'):
            logger.warning(l.strip())

        # here we assume the logfile is no longer needed at this point -
        # remove it
        os.remove(logfile)
    else:
        logger.warning('output\n%s\n', cmd_output)
        logger.warning('end of output')


def __get_proxy_voname__(cred_req):
    """Check validity of proxy vo"""

    vo = credential_store[cred_req].vo

    logger.debug('voms of credential: %s' % vo)
    return vo


def __get_lfc_host__():
    """Gets the LFC_HOST: from current shell or querying BDII on demand"""
    lfc_host = None

    if 'LFC_HOST' in getShell().env:
        lfc_host = getShell().env['LFC_HOST']

    if not lfc_host:
        lfc_host = __get_default_lfc__()

    return lfc_host


def __get_default_lfc__():
    """Gets the default lfc host from lcg-infosites"""

    output = wrap_lcg_infosites('lfc')

    if output == '':
        return None
    else:
        lfc_list = output.strip().split('\n')
        return lfc_list[0]


def __resolve_no_matching_jobs__(cmd_output):
    """Parsing the glite-wms-job-status log to get the glite jobs which have been removed from the WMS"""

    logfile = __resolve_gridcmd_log_path__(
        '(.*-job-status.*\.log)', cmd_output)

    glite_ids = []

    if logfile:

        re_jid = re.compile(r'^Unable to retrieve the status for: (https://\S+:9000/[0-9A-Za-z_.-]+)\s*$')
        re_key = re.compile(r'^.*(no matching jobs found)\s*$')

        myjid = ''
        for line in open(logfile, 'r'):
            m_jid = re_jid.match(line)
            if m_jid:
                myjid = m_jid.group(1)

            if myjid:
                m_key = re_key.match(line)
                if m_key:
                    glite_ids.append(myjid)
                    myjid = ''

    return glite_ids


def list_match(jdlpath, cred_req, ce=None):
    """Returns a list of computing elements can run the job"""

    re_ce = re.compile('^\s*\-\s*(\S+:(2119|8443)/\S+)\s*$')

    matched_ces = []

    cmd = 'glite-wms-job-list-match -a'

    submit_opt = __set_submit_option__()

    if not submit_opt:
        return matched_ces
    else:
        cmd += submit_opt

    cmd = '%s --noint "%s"' % (cmd, jdlpath)

    logger.debug('job list match command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd, allowed_exit=[0, 255])

    for l in output.split('\n'):

        matches = re_ce.match(l)

        if matches:
            matched_ces.append(matches.group(1))

    if ce:
        if matched_ces.count(ce) > 0:
            matched_ces = [ce]
        else:
            matched_ces = []

    logger.debug('== matched CEs ==')
    for myce in matched_ces:
        logger.debug(myce)
    logger.debug('== matched CEs ==')

    return matched_ces


def submit(jdlpath, cred_req, ce=None, perusable=False):
    """Submit a JDL file to LCG"""

    # doing job submission
    cmd = 'glite-wms-job-submit -a'

    submit_opt = __set_submit_option__()

    if not submit_opt:
        return
    else:
        cmd += submit_opt

    if ce:
        cmd += ' -r %s' % ce

    cmd = '%s --nomsg "%s" < /dev/null' % (cmd, jdlpath)

    logger.debug('job submit command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd,
                                            allowed_exit=[0, 255],
                                            timeout=config['SubmissionTimeout'])

    if output:
        output = "%s" % output.strip()

    match = re.search('.*(https://\S+:9000/[0-9A-Za-z_\.\-]+)', output)

    if match:
        logger.debug('job id: %s' % match.group(1))
        if perusable:
            logger.info("Enabling perusal")
            getShell(cred_req).cmd1("glite-wms-job-perusal --set -f stdout %s" % match.group(1))

        # remove the glite command log if it exists
        __clean_gridcmd_log__('(.*-job-submit.*\.log)', output)
        return match.group(1)

    else:
        logger.warning('Job submission failed.')
        __print_gridcmd_log__('(.*-job-submit.*\.log)', output)
        return


def native_master_cancel(jobids, cred_req):
    """Native bulk cancellation supported by GLITE middleware."""

    cmd = 'glite-wms-job-cancel'

    if not __set_submit_option__():
        return False

    idsfile = tempfile.mktemp('.jids')
    with open(idsfile, 'w') as ids_file:
        ids_file.write('\n'.join(jobids) + '\n')

    cmd = '%s --noint -i %s' % (cmd, idsfile)

    logger.debug('job cancel command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd, allowed_exit=[0, 255])

    # clean up tempfile
    if os.path.exists(idsfile):
        os.remove(idsfile)

    if rc != 0:
        logger.warning('Job cancellation failed.')
        __print_gridcmd_log__('(.*-job-cancel.*\.log)', output)
        return False
    else:
        # job cancellation succeeded, try to remove the glite command
        # logfile if it exists
        __clean_gridcmd_log__('(.*-job-cancel.*\.log)', output)
        return True


def status(jobids, cred_req, is_collection=False):
    """Query the status of jobs on the grid"""

    if not jobids:
        return [], []

    idsfile = tempfile.mktemp('.jids')
    with open(idsfile, 'w') as ids_file:
        ids_file.write('\n'.join(jobids) + '\n')

    cmd = 'glite-wms-job-status'

    if is_collection:
        cmd = '%s -v 3' % cmd

    cmd = '%s --noint -i %s' % (cmd, idsfile)
    logger.debug('job status command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd,
                                            allowed_exit=[0, 255],
                                            timeout=config['StatusPollingTimeout'])
    os.remove(idsfile)

    missing_glite_jids = []
    if rc != 0:
        missing_glite_jids = __resolve_no_matching_jobs__(output)

        if missing_glite_jids:
            logger.info(
                'some jobs removed from WMS, will set corresponding Ganga job to \'failed\' status')
            logger.debug('jobs removed from WMS: %s' %
                         repr(missing_glite_jids))
        else:
            __print_gridcmd_log__('(.*-job-status.*\.log)', output)

    # job status query succeeded, try to remove the glite command logfile
    # if it exists
    __clean_gridcmd_log__('(.*-job-status.*\.log)', output)

    re_id = re.compile('^\s*Status info for the Job : (https://.*\S)\s*$')
    re_status = re.compile('^\s*Current Status:\s+(.*\S)\s*$')

    # from glite UI version 1.5.14, the attribute 'Node Name:' is no longer available
    # for distinguishing master and node jobs. A new way has to be applied.
    re_exit = re.compile('^\s*Exit code:\s+(.*\S)\s*$')
    re_reason = re.compile('^\s*Status Reason:\s+(.*\S)\s*$')
    re_dest = re.compile('^\s*Destination:\s+(.*\S)\s*$')

    # pattern to distinguish master and node jobs
    re_master = re.compile('^BOOKKEEPING INFORMATION:\s*$')
    re_node = re.compile('^- Nodes information.*\s*$')

    # pattern for node jobs
    re_nodename = re.compile('^\s*NodeName\s*=\s*"(gsj_[0-9]+)";\s*$')

    info = []
    is_node = False

    for line in output.split('\n'):

        match = re_master.match(line)
        if match:
            is_node = False
            continue

        match = re_node.match(line)
        if match:
            is_node = True
            continue

        match = re_id.match(line)
        if match:
            info += [{'id': match.group(1),
                      'name': '',
                      'is_node': False,
                      'status': '',
                      'exit': '',
                      'reason': '',
                      'destination': ''}]
            if is_node:
                info[-1]['is_node'] = True
            continue

        match = re_nodename.match(line)
        if match and is_node:
            info[-1]['name'] = match.group(1)
            continue

        match = re_status.match(line)
        if match:
            info[-1]['status'] = match.group(1)
            continue

        match = re_exit.match(line)
        if match:
            info[-1]['exit'] = match.group(1)
            continue

        match = re_reason.match(line)
        if match:
            info[-1]['reason'] = match.group(1)
            continue

        match = re_dest.match(line)
        if match:
            info[-1]['destination'] = match.group(1)
            continue

    return info, missing_glite_jids


def get_loginfo(jobids, directory, cred_req, verbosity=1):
    """Fetch the logging info of the given job and save the output in the job's outputdir"""

    cmd = 'glite-wms-job-logging-info -v %d' % verbosity

    log_output = directory + '/__jobloginfo__.log'

    idsfile = tempfile.mktemp('.jids')
    with open(idsfile, 'w') as ids_file:
        ids_file.write('\n'.join(jobids) + '\n')

    cmd = '%s --noint -o %s -i %s' % (cmd, log_output, idsfile)

    logger.debug('job logging info command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd, allowed_exit=[0, 255])
    os.remove(idsfile)

    if rc != 0:
        __print_gridcmd_log__('(.*-logging-info.*\.log)', output)
        return False
    else:
        # logging-info checking succeeded, try to remove the glite command
        # logfile if it exists
        __clean_gridcmd_log__('(.*-logging-info.*\.log)', output)
        # returns the path to the saved logging info if success
        return log_output


def get_output(jobid, directory, cred_req):
    """Retrieve the output of a job on the grid"""

    cmd = 'glite-wms-job-output'
    # general WMS options (somehow used by the glite-wms-job-output
    # command)
    if config['Config']:
        cmd += ' --config %s' % config['Config']

    cmd = '%s --noint --dir %s %s' % (cmd, directory, jobid)

    logger.debug('job get output command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd, allowed_exit=[0, 255])

    match = re.search('directory:\n\s*([^\t\n\r\f\v]+)\s*\n', output)

    if not match:
        logger.warning('Job output fetch failed.')
        __print_gridcmd_log__('(.*-output.*\.log)', output)
        return False, 'cannot fetch job output'

    # job output fetching succeeded, try to remove the glite command
    # logfile if it exists
    __clean_gridcmd_log__('(.*-output.*\.log)', output)

    outdir = match.group(1)

#       some versions of LCG middleware create an extra output directory (named <uid>_<jid_hash>)
#       inside the job.outputdir. Try to match the jid_hash in the outdir. Do output movement
#       if the <jid_hash> is found in the path of outdir.
    import urlparse
    jid_hash = urlparse.urlparse(jobid)[2][1:]

    if outdir.count(jid_hash):
        if getShell(cred_req).system('mv "%s"/* "%s"' % (outdir, directory)) == 0:
            try:
                os.rmdir(outdir)
            except Exception as msg:
                logger.warning(
                    "Error trying to remove the empty directory %s:\n%s" % (outdir, msg))
        else:
            logger.warning("Error moving output from %s to %s.\nOutput is left in %s." % (
                outdir, directory, outdir))
    else:
        pass

    return __get_app_exitcode__(directory)


def cancel_multiple(jobids, cred_req):
    """Cancel multiple jobs in one LCG job cancellation call"""

    # compose a temporary file with job ids in it
    if not jobids:
        return True

    # do the cancellation using a proper LCG command
    cmd = 'glite-wms-job-cancel'

    idsfile = tempfile.mktemp('.jids')
    with open(idsfile, 'w') as ids_file:
        ids_file.write('\n'.join(jobids) + '\n')

    # compose the cancel command
    cmd = '%s --noint -i %s' % (cmd, idsfile)

    logger.debug('job cancel command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd, allowed_exit=[0, 255])

    # clean up tempfile
    if os.path.exists(idsfile):
        os.remove(idsfile)

    if rc == 0:
        # job cancelling succeeded, try to remove the glite command logfile
        # if it exists
        __clean_gridcmd_log__('(.*-job-cancel.*\.log)', output)
        return True
    else:
        logger.warning("Failed to cancel jobs.\n%s" % output)
        __print_gridcmd_log__('(.*-job-cancel.*\.log)', output)
        return False


def cancel(jobid, cred_req):
    """Cancel a job"""

    cmd = 'glite-wms-job-cancel'

    cmd = '%s --noint %s' % (cmd, jobid)

    logger.debug('job cancel command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd, allowed_exit=[0, 255])

    if rc == 0:
        # job cancelling succeeded, try to remove the glite command logfile
        # if it exists
        __clean_gridcmd_log__('(.*-job-cancel.*\.log)', output)
        return True
    else:
        logger.warning("Failed to cancel job %s.\n%s" % (jobid, output))
        __print_gridcmd_log__('(.*-job-cancel.*\.log)', output)
        return False


def __cream_parse_job_status__(log):
    """Parsing job status report from CREAM CE status query"""

    job_info_dict = {}

    re_jid = re.compile(r'^\s+JobID=\[(https://.*[:0-9]?/CREAM.*)\]$')
    re_log = re.compile(r'^\s+(\S+.*\S+)\s+=\s+\[(.*)\]$')

    re_jts = re.compile(r'^\s+Job status changes:$')
    re_ts = re.compile(r'^\s+Status\s+=\s+\[(.*)\]\s+\-\s+\[(.*)\]\s+\(([0-9]+)\)$')
    re_cmd = re.compile(r'^\s+Issued Commands:$')

    # in case of status retrieval failed
    re_jnf = re.compile(r'^.*job not found.*$')

    jid = None

    for jlog in log.split('******')[1:]:

        for l in jlog.split('\n'):
            l.strip()

            m = re_jid.match(l)

            if m:
                jid = m.group(1)
                job_info_dict[jid] = {}
                continue

            if re_jnf.match(l):
                break

            m = re_log.match(l)
            if m:
                att = m.group(1)
                val = m.group(2)
                job_info_dict[jid][att] = val
                continue

            if re_jts.match(l):
                job_info_dict[jid]['Timestamps'] = {}
                continue

            m = re_ts.match(l)
            if m:
                s = m.group(1)
                t = int(m.group(3))
                job_info_dict[jid]['Timestamps'][s] = t
                continue

            if re_cmd.match(l):
                break

    return job_info_dict


def cream_proxy_delegation(ce, delid, cred_req):
    """CREAM CE proxy delegation"""

    if not ce:
        logger.warning('No CREAM CE endpoint specified')
        return

    if not delid:

        logger.debug('making new proxy delegation to %s' % ce)

        cmd = 'glite-ce-delegate-proxy'

        cmd += ' -e %s' % ce.split('/cream')[0]

        delid = '%s_%s' % (credential_store[cred_req].identity, get_uuid())

        cmd = '%s "%s"' % (cmd, delid)

        logger.debug('proxy delegation command: %s' % cmd)

        rc, output, m = getShell(cred_req).cmd1(cmd,
                                                allowed_exit=[0, 255],
                                                timeout=config['SubmissionTimeout'])
        if rc != 0:
            # failed to delegate proxy
            logger.error('proxy delegation error: %s' % output)
            delid = ''
        else:
            # proxy delegated successfully
            t_expire = datetime.datetime.now() + credential_store[cred_req].time_left()

            logger.debug('new proxy at %s valid until %s' % (ce, t_expire))

    return delid


def cream_submit(jdlpath, ce, delid, cred_req):
    """CREAM CE direct job submission"""

    if not ce:
        logger.warning('No CREAM CE endpoint specified')
        return

    cmd = 'glite-ce-job-submit'

    delid = cream_proxy_delegation(ce, delid, cred_req)

    if delid:
        cmd += ' -D "%s"' % delid
    else:
        cmd += ' -a'

    cmd += ' -r %s' % ce

    cmd = '%s --nomsg "%s" < /dev/null' % (cmd, jdlpath)

    logger.debug('job submit command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd,
                                            allowed_exit=[0, 255],
                                            timeout=config['SubmissionTimeout'])

    if output:
        output = "%s" % output.strip()

    match = re.search(r'^(https://\S+:8443/[0-9A-Za-z_\.\-]+)$', output)

    if match:
        logger.debug('job id: %s' % match.group(1))
        return match.group(1)
    else:
        logger.warning('Job submission failed.')
        return


def cream_status(jobids, cred_req):
    """CREAM CE job status query"""

    if not jobids:
        return [], []

    idsfile = tempfile.mktemp('.jids')
    with open(idsfile, 'w') as ids_file:
        ids_file.write('##CREAMJOBS##\n' + '\n'.join(jobids) + '\n')

    cmd = 'glite-ce-job-status'

    cmd = '%s -L 2 -n -i %s' % (cmd, idsfile)
    logger.debug('job status command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd,
                                            allowed_exit=[0, 255],
                                            timeout=config['StatusPollingTimeout'])
    job_info_dict = {}
    if rc == 0 and output:
        job_info_dict = __cream_parse_job_status__(output)

    # clean up tempfile
    if os.path.exists(idsfile):
        os.remove(idsfile)

    return job_info_dict


def cream_purge_multiple(jobids, cred_req):
    """CREAM CE job purging"""

    idsfile = tempfile.mktemp('.jids')
    with open(idsfile, 'w') as ids_file:
        ids_file.write('##CREAMJOBS##\n' + '\n'.join(jobids) + '\n')

    cmd = 'glite-ce-job-purge'

    cmd = '%s -n -N -i %s' % (cmd, idsfile)

    logger.debug('job purge command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd, allowed_exit=[0, 255])

    logger.debug(output)

    # clean up tempfile
    if os.path.exists(idsfile):
        os.remove(idsfile)

    if rc == 0:
        return True
    else:
        return False


def cream_cancel_multiple(jobids, cred_req):
    """CREAM CE job cancelling"""

    idsfile = tempfile.mktemp('.jids')
    with open(idsfile, 'w') as ids_file:
        ids_file.write('##CREAMJOBS##\n' + '\n'.join(jobids) + '\n')

    cmd = 'glite-ce-job-cancel'

    cmd = '%s -n -N -i %s' % (cmd, idsfile)

    logger.debug('job cancel command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd, allowed_exit=[0, 255])

    logger.debug(output)

    # clean up tempfile
    if os.path.exists(idsfile):
        os.remove(idsfile)

    if rc == 0:
        return True
    else:
        return False


def cream_get_output(osb_uri_list, directory, cred_req):
    """CREAM CE job output retrieval"""

    gfiles = []
    for uri in osb_uri_list:
        gf = GridftpFileIndex()
        gf.id = uri
        gfiles.append(gf)

    cache = GridftpSandboxCache()
    cache.uploaded_files = gfiles

    return cache.download(cred_req=cred_req, files=map(lambda x: x.id, gfiles), dest_dir=directory)


def __get_app_exitcode__(outputdir):
    import Ganga.Core.Sandbox as Sandbox

    Sandbox.getPackedOutputSandbox(outputdir, outputdir)

    # check the application exit code
    app_exitcode = -1
    runtime_log = os.path.join(outputdir, '__jobscript__.log')
    pat = re.compile(r'.*exit code (\d+).')

    if not os.path.exists(runtime_log):
        logger.warning('job runtime log not found: %s' % runtime_log)
        return False, 'job runtime log not found: %s' % runtime_log

    for line in open(runtime_log, 'r'):
        mat = pat.match(line)
        if mat:
            app_exitcode = eval(mat.groups()[0])
            break

    # returns False if the exit code of the real executable is not zero
    # the job status of GANGA will be changed to 'failed' if the return
    # value is False
    if app_exitcode != 0:
        logger.debug(
            'job\'s executable returns non-zero exit code: %d' % app_exitcode)
        return False, app_exitcode
    else:
        return True, 0


def expandxrsl(items):
    """Expand xrsl items"""

    xrsl = "&\n"
    for key, value in items.iteritems():

        if key == "inputFiles":
            # special case for input files
            xrsl += "(inputFiles="

            for f in value:
                xrsl += "(\"%s\" \"%s\")\n" % (os.path.basename(f), f)

            xrsl += ")\n"

        elif key == "outputFiles":
            # special case for input files
            xrsl += "(outputFiles="

            for f in value:
                xrsl += "(\"%s\" \"\")\n" % (os.path.basename(f))

            xrsl += ")\n"

        elif isinstance(value, dict):
            # expand if a dictionary
            xrsl += "(%s=" % key
            for key2, value2 in value.iteritems():
                xrsl += "(\"%s\" \"%s\")\n" % (key2, value2)

            xrsl += ")\n"
        else:
            # straight key pair
            xrsl += "(%s=\"%s\")\n" % (key, value)

    return xrsl


def expandjdl(items):
    """Expand jdl items"""

    text = "[\n"
    for key, value in items.iteritems():

        if key == 'Requirements':
            if value:
                text += 'Requirements = \n   %s;\n' % ' &&\n   '.join(
                    value)

        elif key in ['ShallowRetryCount', 'RetryCount', 'NodeNumber', 'ExpiryTime', 'PerusalTimeInterval']:
            try:
                value = int(value)
                if value < 0:
                    raise ValueError
                text += '%s = %d;\n' % (key, value)
            except ValueError:
                logger.warning('%s is not positive integer.' % key)

        elif key == 'Environment':
            if value:
                text += 'Environment = {\n   "%s"\n};\n' % '",\n   "'.join(
                    ['%s=\'%s\'' % var for var in value.items()])

        elif key == 'DataRequirements':
            text += 'DataRequirements = {\n'
            for entry in value:
                text += '  [\n'
                text += '    InputData = {\n'
                for datafile in entry['InputData']:
                    text += '      "%s",\n' % datafile
                # Get rid of trailing comma
                text = text.rstrip(',\n') + '\n'
                text += '    };\n'
                text += '    DataCatalogType = "%s";\n' % entry[
                    'DataCatalogType']
                if 'DataCatalog' in entry:
                    text += '    DataCatalog = "%s";\n' % entry[
                        'DataCatalog']
                text += '  ],\n'
            text = text.rstrip(',\n') + '\n'  # Get rid of trailing comma
            text += '};\n'

        elif isinstance(value, list):
            if value:
                text += '%s = {\n   "%s"\n};\n' % (key,
                                                   '",\n   "'.join(value))

        elif key == 'Rank':
            text += 'Rank = ( %s );\n' % value

        elif key == 'Nodes':
            text += 'Nodes = %s;\n' % value

        elif key in ['PerusalFileEnable', 'AllowZippedISB']:
            text += '%s = %s;\n' % (key, value)

        else:
            text += '%s = "%s";\n' % (key, value)

    text += "\n]\n"
    return text


def wrap_lcg_infosites(opts=""):
    """Wrap the lcg-infosites command"""

    cmd = 'lcg-infosites --vo %s %s' % (
        config['VirtualOrganisation'], opts)

    logger.debug('lcg-infosites command: %s' % cmd)

    rc, output, m = getShell().cmd1('%s' % cmd, allowed_exit=[0, 255])

    if rc != 0:
        return ""
    else:
        return output


def __arc_get_config_file_arg__():
    """Helper function to return the config file argument"""
    if config['ArcConfigFile']:
        return "-z " + config['ArcConfigFile']

    return ""


def arc_submit(jdlpath, ce, verbose, cred_req):
    """ARC CE direct job submission"""

    # No longer need to specify CE if available in client.conf
    # if not ce:
    #    logger.warning('No CREAM CE endpoint specified')
    #    return

    # write to a temporary XML file as otherwise can't submit in parallel
    tmpstr = '/tmp/' + randomString() + '.arcsub.xml'
    cmd = 'arcsub %s -S org.nordugrid.gridftpjob -j %s' % (__arc_get_config_file_arg__(), tmpstr)

    if verbose:
        cmd += ' -d DEBUG '

    if ce:
        cmd += ' -c %s' % ce

    cmd = '%s "%s" < /dev/null' % (cmd, jdlpath)

    logger.debug('job submit command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd,
                                            allowed_exit=[0, 255],
                                            timeout=config['SubmissionTimeout'])

    if output:
        output = "%s" % output.strip()
    getShell().system('rm ' + tmpstr)

    # Job submitted with jobid:
    # gsiftp://lcgce01.phy.bris.ac.uk:2811/jobs/vSoLDmvvEljnvnizHq7yZUKmABFKDmABFKDmCTGKDmABFKDmfN955m
    match = re.search(r'(gsiftp://\S+:2811/jobs/[0-9A-Za-z_\.\-]+)$', output)

    # Job submitted with jobid: https://ce2.dur.scotgrid.ac.uk:8443/arex/..
    if not match:
        match = re.search(r'(https://\S+:8443/arex/[0-9A-Za-z_\.\-]+)$', output)

    if match:
        logger.debug('job id: %s' % match.group(1))
        return match.group(1)
    else:
        logger.warning('Job submission failed.')
        return


def arc_status(jobids, ce_list, cred_req):
    """ARC CE job status query"""

    if not jobids:
        return [], []

    idsfile = tempfile.mktemp('.jids')
    with open(idsfile, 'w') as ids_file:
        ids_file.write('\n'.join(jobids) + '\n')

    cmd = 'arcstat'

    cmd += ' %s -i %s -j %s' % (__arc_get_config_file_arg__(), idsfile, config["ArcJobListFile"])
    logger.debug('job status command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd,
                                            allowed_exit=[0, 1, 255],
                                            timeout=config['StatusPollingTimeout'])
    job_info_dict = {}

    if rc != 0:
        logger.warning('jobs not found in XML file: arcsync will be executed to update the job information')
        __arc_sync__(ce_list, cred_req)

    if rc == 0 and output:
        job_info_dict = __arc_parse_job_status__(output)

    return job_info_dict


def __arc_parse_job_status__(log):
    """Parsing job status report from CREAM CE status query"""

    # Job: gsiftp://lcgce01.phy.bris.ac.uk:2811/jobs/FowMDmswEljnvnizHq7yZUKmABFKDmABFKDmxbGKDmABFKDmlw9pKo
    # State: Finished (FINISHED)
    # Exit Code: 0

    # Job: https://ce2.dur.scotgrid.ac.uk:8443/arex/jNxMDmXTj7jnVDJaVq17x81mABFKDmABFKDmhfRKDmjBFKDmLaCRVn
    # State: Finished (terminal:client-stageout-possible)
    # Exit Code: 0

    job_info_dict = {}
    jid = None

    for ln in log.split('\n'):

        ln.strip()

        # do we have a failed retrieval?
        if ln.find("Job not found") != -1:
            logger.warning("Could not find info for job id '%s'" % jid)
            jid = None
        elif ln.find("Job:") != -1 and ln.find("gsiftp") != -1:
            # new job info block
            jid = ln[ln.find("gsiftp"):].strip()
            job_info_dict[jid] = {}
        elif ln.find("Job:") != -1 and ln.find("https") != -1:
            # new job info block
            jid = ln[ln.find("https"):].strip()
            job_info_dict[jid] = {}

        # get info
        if ln.find("State:") != -1:
            job_info_dict[jid]['State'] = ln[ln.find("State:") + len("State:"):].strip()

        if ln.find("Exit Code:") != -1:
            job_info_dict[jid]['Exit Code'] = ln[ln.find("Exit Code:") + len("Exit Code:"):].strip()

        if ln.find("Job Error:") != -1:
            job_info_dict[jid]['Job Error'] = ln[ln.find("Job Error:") + len("Job Error:"):].strip()

    return job_info_dict


def __arc_sync__(ce_list, cred_req):
    """Collect jobs to jobs.xml"""

    if ce_list[0]:
        cmd = 'arcsync %s -j %s -f -c h%s' % (__arc_get_config_file_arg__(
        ), config["ArcJobListFile"], ' -c '.join(ce_list))
    else:
        cmd = 'arcsync %s -j %s -f ' % (
            __arc_get_config_file_arg__(), config["ArcJobListFile"])

    logger.debug('sync ARC jobs list with: %s' % cmd)
    rc, output, m = getShell(cred_req).cmd1(cmd,
                                            allowed_exit=[0, 255],
                                            timeout=config['StatusPollingTimeout'])

    if rc != 0:
        logger.error('Unable to sync ARC jobs. Error: %s' % output)


def arc_get_output(jid, directory, cred_req):
    """ARC CE job output retrieval"""

    # construct URI list from ID and output from arcls
    cmd = 'arcls %s %s' % (__arc_get_config_file_arg__(), jid)
    logger.debug('arcls command: %s' % cmd)
    rc, output, m = getShell(cred_req).cmd1(cmd,
                                            allowed_exit=[0, 255],
                                            timeout=config['SubmissionTimeout'])
    if rc:
        logger.error(
            "Could not find directory associated with ARC job ID '%s'" % jid)
        return False

    # URI is JID + filename
    gfiles = []
    for uri in output.split("\n"):
        if len(uri) == 0:
            continue
        uri = jid + "/" + uri
        gf = GridftpFileIndex()
        gf.id = uri
        gfiles.append(gf)

    cache = GridftpSandboxCache()
    cache.uploaded_files = gfiles
    return cache.download(cred_req=cred_req, files=map(lambda x: x.id, gfiles), dest_dir=directory)


def arc_purge_multiple(jobids, cred_req):
    """ARC CE job purging"""

    idsfile = tempfile.mktemp('.jids')
    with open(idsfile, 'w') as ids_file:
        ids_file.write('\n'.join(jobids) + '\n')

    cmd = 'arcclean'

    cmd = '%s %s -i %s -j %s' % (
        cmd, __arc_get_config_file_arg__(), idsfile, config["ArcJobListFile"])

    logger.debug('job purge command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd, allowed_exit=[0, 255])

    logger.debug(output)

    if rc == 0:
        return True
    else:
        return False


def arc_cancel(jobid, cred_req):
    """Cancel a job"""

    cmd = 'arckill'

    cmd = '%s %s %s -j %s' % (cmd, str(
        jobid)[1:-1], __arc_get_config_file_arg__(), config["ArcJobListFile"])

    logger.debug('job cancel command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd, allowed_exit=[0, 255])

    if rc == 0:
        # job cancelling succeeded, try to remove the glite command logfile
        # if it exists
        __clean_gridcmd_log__('(.*-job-cancel.*\.log)', output)
        return True
    else:
        logger.warning("Failed to cancel job %s.\n%s" % (jobid, output))
        __print_gridcmd_log__('(.*-job-cancel.*\.log)', output)
        return False


def arc_cancel_multiple(jobids, cred_req):
    """Cancel multiple jobs in one LCG job cancellation call"""

    # compose a temporary file with job ids in it
    if not jobids:
        return True

    cmd = 'arckill'

    idsfile = tempfile.mktemp('.jids')
    with open(idsfile, 'w') as ids_file:
        ids_file.write('\n'.join(jobids) + '\n')

    # compose the cancel comman
    cmd = '%s %s -i %s -j %s' % (
        cmd, __arc_get_config_file_arg__(), idsfile, config["ArcJobListFile"])

    logger.debug('job cancel command: %s' % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd, allowed_exit=[0, 255])

    if rc == 0:
        # job cancelling succeeded, try to remove the glite command logfile
        # if it exists
        __clean_gridcmd_log__('(.*-job-cancel.*\.log)', output)
        return True
    else:
        logger.warning("Failed to cancel jobs.\n%s" % output)
        __print_gridcmd_log__('(.*-job-cancel.*\.log)', output)
        return False


def arc_info(cred_req):
    """Run the arcinfo command"""

    cmd = 'arcinfo %s > /dev/null' % __arc_get_config_file_arg__()
    logger.debug("Running arcinfo command '%s'" % cmd)

    rc, output, m = getShell(cred_req).cmd1(cmd,
                                            allowed_exit=[0, 1, 255],
                                            timeout=config['StatusPollingTimeout'])
    return rc, output




class PackageSetup(object):

    """ PackageSetup objects represent the external packages required by a given runtime unit of Ganga.
    The package information is stored in a dictionary. See Ganga/PACKAGE.py
    """

    def __init__(self, packages):
        """
        Initialize the Package object with the list of required external packages
        """
        self.packages = packages

    def getPackagePath(self, name, force=False):
        """ Return a tuple (path,tarball) which describes the external package:
             - path points to the package top directory
             - tarball is the package tarball name

            If *force=False* (default) the python version is checked and if the package
            is not required then return ('','').

            If *force=True* the path is returned even if the package is not required.            
        """

        import os.path
        import sys

        package = self.packages[name]

        try:
            package_required = sys.hexversion < int(
                package['maxHexVersion'], 16) or force
        except (KeyError, ValueError):
            package_required = 1

        if not package_required:
            return ('', '')

        platfdir = getPlatform()

        if 'noarch' in package and package['noarch']:
            platfdir = 'noarch'

        return (os.path.join(getExternalHome(), name, package['version'], platfdir),
                '-'.join([name, package['version'], platfdir, 'ganga.tar.gz']))

    def getPackagePath2(self, name, var, force=False):
        """ Return a path to the package for specified environment variable.
        For example: _getPackagePath('x','LD_LIBRARY_PATH') returns a LD_LIBRARY_PATH for the package (if defined).
        The value may be a string or a list of strings - in the later case the individual path components will
        be separated by colons (':').
        If the path is not defined (see getPackagePath()) then return an empty string.
        """
        import os
        path, tarball = self.getPackagePath(name, force)
        if path and var in self.packages[name]:
            from Ganga.Utility.util import isStringLike

            ppath = self.packages[name][var]

            if isStringLike(ppath):
                ppath = [ppath]

            return ':'.join(map(lambda p: os.path.join(path, p), ppath))

# for p in ppath:
##                 result += os.path.join(path,p)
# return result
        else:
            return ''

    def prependPath(self, name, var):
        """ Update environment (os.environ) and prepend a package path to var.
        """
        import sys
        import os.path

        if getExternalHome():
            path = self.getPackagePath2(name, var)
            if path:
                try:
                    pp = ':' + os.environ[var]
                except KeyError:
                    pp = ''
                os.environ[var] = path + pp

    def setSysPath(self, name, set_python_path=True):
        """ Update the sys.path variable for a given package by adding a given package to the PYTHONPATH.
        It assumes that the first item in PYTHONPATH is Ganga, so the package path is added as a second item.
        If set_python_path is True, it will update the PYTHONPATH as well
        """
        import sys
        import os

        if getExternalHome():

            # get the path(s)
            paths = self.getPackagePath2(name, 'syspath')
            if not paths:
                return True

            # loop over the sys paths to be added
            for path in reversed(paths.split(':')):
                if path not in sys.path:

                    # add to sys path
                    sys.path.insert(1, path)

                # add to PYTHONPATH?
                if set_python_path:
                    if 'PYTHONPATH' in os.environ:
                        pp_toks = os.environ['PYTHONPATH'].split(':')
                    else:
                        # PYTHONPATH is empty so add Ganga in first
                        pp_toks = [sys.path[0]]

                    # create a new PYTHONPATH
                    if path not in pp_toks:
                        pp_toks.insert(1, path)

                    os.environ['PYTHONPATH'] = ':'.join(pp_toks)

        return True

    def setPath(self, name, var):
        """ 
        Update environment (os.environ) and *set* (overwrite) a package path to var.
        """
        import sys
        import os

        if getExternalHome():
            path = self.getPackagePath2(name, var)
            if path:
                os.environ[var] = path

        return True


def checkPythonVersion(minVersion, minHexVersion):
    """Function to check that the Python version number is greater
       than the minimum required.

       Arguments:
          minVersion    - String representation of minimum version required
          minHexVersion - Hexadecimal representation of minimum version required

       Return value: True if Python version number is greater than minimum,
                     False otherwise"""

    import sys

    status = True

    if sys.hexversion < minHexVersion:
        from Ganga.Utility.logging import getLogger
        logger = getLogger(modulename=True)
        logger.error("Ganga requires Python version %s or greater" %
                     str(minVersion))
        logger.error("Currently using Python %s" % sys.version.split()[0])
        status = False

    return status


# Override the default guessed platform and external dir
# save the first update request to detect conflicts generated by
# two Ganga packages setting a different platform
_new_platform = None


def setPlatform(platform):
    """
    Override globally the auto-detected platform string
    """
    global _new_platform

    if _new_platform and _new_platform != platform:
        raise RuntimeError(
            "Ganga platform has been already set to: %s" % str(_new_platform))
    # change the global platform used in bootstrap
    import Ganga
    if Ganga.PACKAGE._defaultPlatform != platform:
        from Ganga.Utility.logging import getLogger
        logger = getLogger(modulename=True)
        logger.info('The platform identification string (%s) used to resolve Ganga dependencies has been explicitly set to: %s.'
                    % (Ganga.PACKAGE._defaultPlatform, platform))
    Ganga.PACKAGE._defaultPlatform = _new_platform = platform


def getPlatform():
    """
    Returns the current platform string set for Ganga
    """
    from Ganga.PACKAGE import _defaultPlatform
    return _defaultPlatform


def setExternalHome(externalHome):
    """
    Set the dir hosting the external packages
    """
    import Ganga
    Ganga.PACKAGE._defaultExternalHome = externalHome


def getExternalHome():
    """
    Returns the current external home directory set for Ganga
    """
    from Ganga.PACKAGE import _defaultExternalHome
    return _defaultExternalHome




from Ganga.Core.exceptions import GangaException
from Ganga.GPIDev.Lib.File.File import File
from Ganga.GPIDev.Lib.File.LocalFile import LocalFile

import Ganga.Utility.logging
import fnmatch

import os

logger = Ganga.Utility.logging.getLogger()

transformDictionary = {}

def loadScript(scriptFilePath, indentation):

    if not os.path.exists(scriptFilePath):
        from Ganga.Core.exceptions import GangaException
        raise GangaException("Error Finding script file: %s" % str(scriptFilePath))

    with open(scriptFilePath) as this_file:
        file_data = this_file.read()

    return indentScript(file_data, indentation)

def indentScript(script, indenting):

    script_lines = script.split('\n')
    output_script = []
    for this_line in script_lines:
        output_script.append(str(indenting) + str(this_line))

    return '\n'.join(output_script)

def __populate():
    if len(transformDictionary) == 0:
        transformDictionary[File] = LocalFile
        # transformDictionary[ ] =
__populate()

def safeTransformFile(input_file):

    this_type = type(input_file)
    if this_type in transformDictionary:
        return transformDictionary[this_type](input_file)

    logger.error("Cannot safely transform file of type: %s" % (this_type))

    return None

def doesFileExist( input_file=None, input_list = None ):

    if input_file is None:
        input_file = 'some.file'
    if input_list is None:
        input_list = []

    test_fileName = ''
    if type(input_file) == type(''):
        test_fileName = input_file
    elif hasattr(input_file, 'namePattern'):
        test_fileName = input_file.namePattern
    else:
        raise GangaException("Failed to understand file: %s" % str(input_file))

    have_matched = False
    for input_fileName in input_list:
        check_str = ''
        if type(input_fileName) == type(''):
            check_str = input_fileName
        elif hasattr(input_fileName, 'namePattern'):
            check_str = input_fileName.namePattern
        else:
            raise GangaException("Failed to understand file or pattern: %s" % str(input_fileName))

        if fnmatch.fnmatch( test_fileName, check_str ):
            have_matched = True
            break

    return have_matched




#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

from Ganga.Core.exceptions import GangaException
from Ganga.GPIDev.Lib.Dataset import Dataset
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, GangaFileItem
from Ganga.GPIDev.Base.Proxy import getName
import Ganga.Utility.logging
from Ganga.GPIDev.Adapters.IGangaFile import IGangaFile
logger = Ganga.Utility.logging.getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class GangaDataset(Dataset):

    '''Class for handling generic datasets of input files
    '''
    schema = {}
    docstr = 'List of File objects'
    schema['files'] = GangaFileItem(defvalue=[], sequence=1, doc="list of file objects that will be the inputdata for the job")
    schema['treat_as_inputfiles'] = SimpleItem(defvalue=False, doc="Treat the inputdata as inputfiles, i.e. copy the inputdata to the WN")
    _schema = Schema(Version(3, 0), schema)
    _category = 'datasets'
    _name = "GangaDataset"
    _exportmethods = ['append', 'extend', '__len__', 'isEmtpy', 'getFileNames', 'getFilenameList', '__getitem__', '__nonzero__', 'isEmpty',
                        'getFileNames', 'getFilenameList', 'difference', 'isSubset', 'isSuperset', 'symmetricDifference', 'intersection',
                        'union']

    def __init__(self, files=None):
        if files is None:
            files = []
        super(GangaDataset, self).__init__()
        self.files = files

    def __len__(self):
        """The number of files in the dataset."""
        result = 0
        if self.files:
            result = len(self.files)
        return result

    def __nonzero__(self):
        """This is always True, as with an object."""
        return True

    def __getitem__(self, i):
        '''Proivdes scripting (e.g. ds[2] returns the 3rd file) '''
        if isinstance(i, type(slice(0))):
            ds = GangaDataset(files=self.files[i])
            return ds
        else:
            return self.files[i]

    def isEmpty(self): return not bool(self.files)

    def append(self, input_file):
        self.extend([input_file])

    def extend(self, files, unique=False):
        '''Extend the dataset. If unique, then only add files which are not
        already in the dataset.'''
        from Ganga.GPIDev.Base import ReadOnlyObjectError
        if not hasattr(files, "__getitem__"):
            raise GangaException('Argument "files" must be a iterable.')
        if self._getParent() is not None and self._getParent()._readonly():
            raise ReadOnlyObjectError(
                'object Job#%s  is read-only and attribute "%s/inputdata" cannot be modified now' % (self._getParent().id, getName(self)))
        names = self.getFileNames()
        files = [f for f in files]  # just in case they extend w/ self
        for f in files:
            if unique and f.name in names:
                continue
            self.files.append(f)

    def getFileNames(self):
        'Returns a list of the names of all files stored in the dataset.'
        names = []
        for i in self.files:
            if hasattr(i, 'lfn'):
                names.append(i.lfn)
            else:
                try:
                    names.append(i.namePattern)
                except:
                    logger.warning("Cannot determine filename for: %s " % i)
                    raise GangaException("Cannot Get File Name")

        return names

    def getFilenameList(self):
        "return a list of filenames to be created as input.txt on the WN"
        filelist = []
        for f in self.files:
            if hasattr(f, 'accessURL'):
                filelist += f.accessURL()
            elif hasattr(f, 'getFilenameList'):
                filelist += f.getFilenameList()
            else:
                logger.warning(
                    "accessURL or getFilenameList not implemented for File '%s'" % getName(f))

        return filelist

    def difference(self, other):
        '''Returns a new data set w/ files in this that are not in other.'''
        other_files = other.getFullFileNames()
        files = set(self.getFullFileNames()).difference(other_files)
        data = GangaDataset()
        data.extend([list(files)])
        data.depth = self.depth
        return data

    def isSubset(self, other):
        '''Is every file in this data set in other?'''
        return set(self.getFileNames()).issubset(other.getFileNames())

    def isSuperset(self, other):
        '''Is every file in other in this data set?'''
        return set(self.getFileNames()).issuperset(other.getFileNames())

    def symmetricDifference(self, other):
        '''Returns a new data set w/ files in either this or other but not
        both.'''
        other_files = other.getFullFileNames()
        files = set(self.getFullFileNames()).symmetric_difference(other_files)
        data = GangaDataset()
        data.extend([list(files)])
        data.depth = self.depth
        return data

    def intersection(self, other):
        '''Returns a new data set w/ files common to this and other.'''
        other_files = other.getFullFileNames()
        files = set(self.getFullFileNames()).intersection(other_files)
        data = GangaDataset()
        data.extend([list(files)])
        data.depth = self.depth
        return data

    def union(self, other):
        '''Returns a new data set w/ files from this and other.'''
        files = set(self.getFullFileNames()).union(other.getFullFileNames())
        data = GangaDataset()
        data.extend([list(files)])
        data.depth = self.depth
        return data

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



#!/usr/bin/env python
##################################################
# Toolset to assist with GangaSNOplus RATUser and
# RATProd.
# 
# Author: Matt Mottram <m.mottram@qmul.ac.uk>
#
# - Macro checking tools
# - Git archiving tools
# - DB access and Grid options tools
##################################################

import os
import optparse
import subprocess
import urlparse
import urllib2
import httplib
import fnmatch
import tarfile
import shutil
import base64
import getpass
import re
import getpass
import pickle
import time

from GangaSNOplus.Lib.Applications import job_tools


######################################################################################
# Git archiving tools
#
######################################################################################

def download_snapshot(fork, version, filename, username=None, password=None, retry=False):
    """Download a tarball of a given rat version.

    Version may be either the commit hash or the branch name (if latest commit is desired).
    However, a commit hash is preferred as the branch name will mean that newer commits are not 
    grabbed if rerunning at a later date.
    """
    url = "https://github.com/%s/rat/archive/%s.tar.gz" % (fork, version)
    url_request = urllib2.Request(url)
    # Only ever downloading once, so prompt for the username here
    if not username:
        username = raw_input("Username: ") #This might not be the same as the fork...
    if not password:
        password = getpass.getpass("Password: ")
    b64string = base64.encodestring('%s:%s' % (username, password)).replace('\n', '')
    url_request.add_header("Authorization", "Basic %s" % b64string)
    try:
        remote_file = urllib2.urlopen(url_request)
    except urllib2.URLError as e:
        print "Cannot connect to GitHub: ", e
        raise
    try:
        download_size = int(remote_file.info().getheaders("Content-Length")[0])
    except:
        # For some reason GitHub sometimes rejects the first connection attempt.
        if retry is False:
            download_snapshot(fork, version, filename, username, password, True)
        else:
            raise
    local_file = open(filename, "wb")    
    local_file.write(remote_file.read())
    local_file.close()
    remote_file.close()


def make_rat_snapshot(fork, version, update, zip_prefix='archived/',
                      cache_path=os.path.expanduser('~/gaspCache')):
    '''Create a snapshot of RAT from an existing git repo.

    Stores the snapshot in the cache_path, downloads from the named 
    github fork and version; update=True forces an updated version.
    '''
    if not os.path.exists(cache_path):
        os.makedirs(cache_path)
    rat_path = os.path.join(cache_path, 'rat')
    # Old method: download rat git repo, make an tarball of the required commit
    # New method: download directly from the url
    filename = os.path.join(cache_path, "rat.%s.%s.tar.gz" % (fork, version))
    if update is True or not os.path.exists(filename):
        download_snapshot(fork, version, filename)
    return filename

######################################################################################
# Macro checking tools
#
######################################################################################

def check_command(filename, command):
    '''Check to see if a command is present.
    '''
    f = file(filename, 'r')
    command_exists = False
    for line in f.readlines():
        if check_command_line(command, line):
            command_exists = True
    return command_exists

def check_option(filename, command):
    '''Check to see if a command and an option for that command is present.
    '''
    f = file(filename, 'r')
    option_exists = False
    for line in f.readlines():
        if check_option_line(command, line):
            option_exists = True
    return option_exists

def check_command_line(command, line):
    '''Checks for a list of commands/options.

    Returns True if all are present.
    Separations by whitespace only.
    Only takes lists with 1/2 members currently.
    '''
    pattern = re.compile(r'''\s*(?P<command>\S*)\s*(?P<option>\S*)''')
    search = pattern.search(line)
    parts = ['command', 'option']
    match = []
    for i in range(len(command)):
        match.append(0)
    for i, part in enumerate(parts):
        if i >= len(command):
            continue
        if command[i] == search.group(part):
            match[i] = 1
    full_match = True
    for i in match:
        if i == 0:
            full_match = False
    return full_match

def check_option_line(command, line):
    '''Checks for a command string.

    Returns true if an option is present.
    (Will return False if no command is present or if command but no option). 
    '''
    pattern = re.compile(r'''\s*(?P<command>\S*)\s*(?P<option>\S*)''')
    search = pattern.search(line)
    parts = ['command', 'option']
    has_option = False
    option_parts = parts[1:]
    command_part = parts[0]
    if command == search.group(command_part):
        # The command is present, is there an option?
        for part in option_parts:
            if search.group(part) != '':
                # Option is present
                has_option = True
    return has_option



######################################################################################
# Resource checking functions
#
######################################################################################

def get_ce_set():
    '''Just a call to lcg-infosites ce
    '''
    ce_set = set()
    rtc, out, err = job_tools.execute('lcg-infosites', ['--vo', 'snoplus.snolab.ca', 'ce'])
    for line in out:
        bits = line.split()
        if len(bits)==6:
            ce_name = urlparse.urlparse("ce://%s" % bits[5]) # fake a scheme (ce://) for urlparse
            ce_set.add(unicode(ce_name.hostname)) # unicode for simpler comparison with database
    return ce_set


######################################################################################
# Database functions
#
######################################################################################

def encode_url(db_host, relative_path, query_options = None):
    '''Set up url for couchdb query

    db_host should be a urlparse.ParseResult object.
    '''
    db_url = urlparse.urlunparse(db_host)
    url = urlparse.urljoin(db_url, relative_path)
    if query_options is not None and len(query_options):
        query_string = urllib.urlencode(query_options, True)
        url = "%s?%s" % (url, query_string)
    return url


def get_response(host, port, url, request_type = "GET", headers = None, body = None):
    import json # Don't import at the top, python 2.5+ required
    if port is not None:
        connection = httplib.HTTPConnection(host, port=port)
    else:
        connection = httplib.HTTPConnection(host)
    try:
        connection.request(request_type, url, body=body, headers=headers)
        response = connection.getresponse()
    except httplib.HTTPException as e:
        sys.stderr.write('Error accessing the requested db query: %s' % str(e))
        sys.exit(20)
    return json.loads(response.read())


######################################################################################
# .gangasnoplus configuration
#
######################################################################################

class GridConfig:
    '''A singleton object to cache information on grid configuration information.
    '''
    
    _instance = None

    class SingletonHelper:
        def __call__(self, *args, **kw):
            if GridConfig._instance is None:
                object = GridConfig()
                GridConfig._instance = object
            return GridConfig._instance

    get_instance = SingletonHelper()

    def __init__(self):
        '''Initialise and load configuration access information.
        '''
        self._last_update = None # To ensure the first call updates
        self._refresh_every = 3600 # Update once per hour
        self._worker_node_info = {}
        self._config_path = os.path.expanduser("~/.gangasnoplus")
        self._config_version = 1

    def load_access(self):
        '''Load access credentials for the processing database
        
        If unavailable, create them.
        '''
        try:
            cred_file = file(self._config_path, 'r')
            info = pickle.load(cred_file)
            cred_file.close()
            if 'version' not in info or \
               info['version'] < self._config_version:
                self.set_access()
                return self.load_access()
            return info
        except IOError:
            # Loading for the first time
            self.set_access()
            if self.load_access() is None:
                raise
        return None

    def set_access(self):
        '''Create the access credentials.
        '''
        cred_file = file(self._config_path, 'w')
        server = raw_input("Set processing database URL: ")
        name = raw_input("Set processing database name: ")
        username = raw_input("Set database username: ")
        password = getpass.getpass("Set database password: ")
        if not urlparse.urlparse(server).hostname:
            server = 'http://' + server # so that urlparse works
        url = urlparse.urlparse(server) # save as a ParseResult class 
        info = {"url": url, "name": name, "credentials": base64.encodestring('%s:%s' % (username, password))[:-1],
                "version": self._config_version}
        pickle.dump(info, cred_file)
        cred_file.close()

    def get_worker_node_info(self):
        '''Return information on worker node statuses.
        '''
        if not self._last_update or \
           time.time() > (self._last_update + self._refresh_every):
            self.refresh_worker_node_info()
            self._last_update = time.time()
        return self._worker_node_info

    def get_excluded_worker_nodes(self):
        '''Just return the CE sites which should be excluded.
        '''
        exclude_list = []
        wn_dict = self.get_worker_node_info()
        for wn in wn_dict:
            if wn_dict[wn] is False:
                exclude_list.append(wn)
        return exclude_list

    def refresh_worker_node_info(self):
        '''Update WN information from database.
        '''
        database_info = self.load_access()
        url = database_info['url']
        query_url = encode_url(url, "%s/_design/ganga/_view/ce_list" % (database_info['name']))
        response = get_response(url.hostname, url.port, query_url, "GET",
                                {"Authorization": "Basic %s" % database_info['credentials']})
        results = response["rows"]
        database_ce_info = {}
        for i, row in enumerate(results):
            database_ce_info[row['key']] = row['value']
        ce_set = get_ce_set()
        self._worker_node_info = {}
        for ce in ce_set:
            if ce not in database_ce_info:
                print "Warning: missing information for %s, exclude" % ce
                self._worker_node_info[ce] = False
            else:
                self._worker_node_info[ce] = database_ce_info[ce]



from __future__ import absolute_import

import copy
import os
from abc import ABCMeta, abstractmethod
from datetime import datetime, timedelta
from functools import wraps

import Ganga.Utility.logging

from Ganga.Core.exceptions import CredentialsError
from Ganga.GPIDev.Credentials.CredentialStore import credential_store

logger = Ganga.Utility.logging.getLogger()

def cache(method):
    """
    The cache decorator can be applied to any method in an ``ICredentialInfo` subclass.
    It stores the return value of the function in the ``self.cache`` dictionary
    with the key being the name of the function.
    The cache is invalidated if ``os.path.getmtime(self.location)`` returns
    greater than ``self.cache['mtime']``, i.e. if the file has changed on disk

    Not having to call the external commands comes at the cost of calling ``stat()``
    every time a parameter is accessed but this is still a saving of many orders of
    magnitude.

    Args:
        method (function): This is the method which we're wrapping
    """
    @wraps(method)
    def cache_function(self, *args, **kwargs):

        # If the mtime has been changed, clear the cache
        if credential_store.enable_caching and os.path.exists(self.location):
            mtime = os.path.getmtime(self.location)
            if mtime > self.cache['mtime']:
                self.cache = {'mtime': mtime}
        else:
            self.cache = {'mtime': 0}

        # If entry is missing from cache, repopulate it
        # This will run if the cache was just cleared
        if method.func_name not in self.cache:
            self.cache[method.func_name] = method(self, *args, **kwargs)

        return self.cache[method.func_name]
    return cache_function

def retry_command(method):
    """
    This method attempts to run the same function which can exit due to a CredetialsError
    Any other exceptions are raised as appropriate. If a command fails it's assumed to have failed due to an invalid user input and is retried upto crednential_store.retry_limit
    Args:
        method (function): This is the method which we're wrapping
    """
    @wraps(method)
    def retry_function(*args, **kwds):
        for _ in range(credential_store.retry_limit):
            try:
                return method(*args, **kwds)
            except CredentialsError:
                # We expect errors of this type
                pass
        return None
    return retry_function

class ICredentialInfo(object):
    """
    The interface for all credential types.
    Each object covers one credential file exactly.
    The credential file is central to the object and all information is gathered from there.

    These are only created by the store and should not be persisted.
    """
    __metaclass__ = ABCMeta

    def __init__(self, requirements, check_file=False, create=False):
        # type: (ICredentialRequirement, bool, bool) -> None
        """
        Args:
            requirements (ICredentialRequirement): An object specifying the requirements
            check_file (bool): Raise an exception if the file does not exist
            create (bool): Create the credential file

        Raises:
            IOError: If told to wrap a non-existent file
            CredentialsError: If this object cannot satisfy ``requirements``
        """
        super(ICredentialInfo, self).__init__()

        self.cache = {'mtime': 0}

        self.initial_requirements = copy.deepcopy(requirements)  # Store the requirements that the object was created with. Used for creation

        if check_file:
            logger.debug('Trying to wrap %s', self.location)
            if not self.exists():
                raise IOError('Proxy file {path} not found'.format(path=self.location))
            logger.debug('Wrapping existing file %s', self.location)

        if create:
            logger.debug('Making a new one')
            self.create()

        # If the proxy object does not satisfy the requirements then abort the construction
        if not self.check_requirements(requirements):
            raise CredentialsError('Proxy object cannot satisfy its own requirements')

    def __str__(self):
        """ Returns a user-readable string describing the credential and it's validity """
        return '{class_name} at {file_path} : TimeLeft = {time_left}, Valid = {currently_valid}'.format(\
                        class_name=type(self).__name__, file_path=self.location, currently_valid=self.is_valid(), time_left=self.time_left())

    def _repr_pretty_(self, p, cycle):
        """ A wrapper to self.__str__ for the pretty print methods in IPython """
        p.text(str(self))

    @property
    def location(self):
        # type: () -> str
        """
        The location of the file on disk
        """
        location = self.default_location()
        encoded_ext = self.initial_requirements.encoded()
        if encoded_ext and not location.endswith(encoded_ext):
            location += ':' + encoded_ext

        return location

    @abstractmethod
    def default_location(self):
        # type: () -> str
        """
        Returns the default location for the credential file.
        This is the location that most tools will look for the file
        or where the file is created without specifying anything.
        """
        pass

    @abstractmethod
    def create(self):
        # type: () -> None
        """
        Create a new credential file
        """
        pass

    def renew(self):
        # type: () -> None
        """
        Renew an existing credential file
        """
        self.create()

    def is_valid(self):
        # type: () -> bool
        """
        Is the credential valid to be used
        """
        # TODO We should check that there's more than some minimum time left on the proxy
        return self.time_left() > timedelta()

    @abstractmethod
    def expiry_time(self):
        # type: () -> datetime.datetime
        """
        Returns the expiry time
        """
        pass

    def time_left(self):
        # type: () -> datetime.timedelta
        """
        Returns the time left
        """
        time_left = self.expiry_time() - datetime.now()
        return max(time_left, timedelta())

    def check_requirements(self, query):
        # type: (ICredentialRequirement) -> bool
        """
        Args:
            query (ICredentialRequirement): The requirements to check ourself against

        Checks all requirements.

        Returns:
            ``True`` if we meet all requirements
            ``False`` if even one requirement is not met or if the credential is not valid
        """
        if not self.exists():
            return False
        return all(self.check_requirement(query, requirementName) for requirementName in query._schema.datadict)

    def check_requirement(self, query, requirement_name):
        # type: (ICredentialRequirement, str) -> bool
        """
        Args:
            query (ICredentialRequirement):
            requirement_name (str): The requirement attribute to check

        Returns:
            ``True`` if ``self`` matches ``query``'s ``requirement``.
        """
        requirement_value = getattr(query, requirement_name)
        if requirement_value is None:
            # If this requirementName is unspecified then ignore it
            return True
        logger.debug('\'%s\': \t%s \t%s', requirement_name, getattr(self, requirement_name), requirement_value)
        return getattr(self, requirement_name) == requirement_value

    def exists(self):
        # type: () -> bool
        """
        Does the credential file exist on disk
        """
        return os.path.exists(self.location)

    def __eq__(self, other):
        """
        Test the equality of a crednetial against another by comparing their on-disk location-s
        Args:
            other (ICredentialInfo) This is the object being compared against
        """
        return self.location == other.location

    def __ne__(self, other):
        """
        Returns the inverse of the equality check
        Args:
            other (ICredentialInfo) This is the object being compared against
        """
        return not self == other

    def __hash__(self):
        """
        Returns a hash of the location of the object on disk
        """
        return hash(self.location)




# Import classes that should be in the Tasks namespace
## Import the list of tasks and the task and abstract job definition

from AtlasTask import AtlasTask
from AtlasTransform import AtlasTransform
from AtlasUnit import AtlasUnit





# Logical operators and helpers.


def implies(p, q):
    return not p or q


def equivalent(p, q):
    return (p and q) or (not p and not q)


def xor(p, q):
    return (p and not q) or (q and not p)



from __future__ import absolute_import
from threading import Thread, RLock
from .GangaThreadPool import GangaThreadPool
from Ganga.Utility.logging import getLogger
import functools
logger = getLogger()

def synchronised(f):
    @functools.wraps(f)
    def decorated(self, *args, **kwargs):
        with self._lock:
            return f(self, *args, **kwargs)
    return decorated


class GangaThread(Thread):

    def __init__(self, name, auto_register=True, critical=True, **kwds):

        self.gangaName = str(name)  # want to copy actual not by ref!
        name = 'GANGA_Update_Thread_%s' % name

        Thread.__init__(self, args=list(), name=name, **kwds)
        self.setDaemon(True)
        self.__should_stop_flag = False
        self.__critical = critical

        if auto_register:
            tpool = GangaThreadPool.getInstance()
            tpool.addServiceThread(self)
        self._lock = RLock()

    @synchronised
    def isCritical(self):
        """Return critical flag.

        @return: Boolean critical flag.
        """
        return self.__critical

    @synchronised
    def setCritical(self, critical):
        """Set critical flag, which can be used for example in shutdown
        algorithms. See Ganga/Core/__init__.py for example.

        @param critical: Boolean critical flag.
        """
        self.__critical = critical

    @synchronised
    def should_stop(self):
        return self.__should_stop_flag

    @synchronised
    def stop(self):
        if not self.__should_stop_flag:
            logger.debug("Stopping: %s", self.gangaName)
            self.__should_stop_flag = True

    @synchronised
    def unregister(self):
        GangaThreadPool.getInstance().delServiceThread(self)

    @synchronised
    def register(self):
        GangaThreadPool.getInstance().addServiceThread(self)



##########################################################################
# Ganga - a computational task management tool for easy access to Grid resources
# http://cern.ch/ganga
#
# $Id: GridShell.py,v 1.1 2008-07-17 16:41:00 moscicki Exp $
#
# Copyright (C) 2003 The Ganga Project
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# LCG backend
#
# ATLAS/ARDA
#
# Date:   August 2006
##########################################################################

import os

from Ganga.Utility.Shell import Shell
from Ganga.Utility.Config import getConfig, ConfigError
from Ganga.Utility.logging import getLogger

from Ganga.GPIDev.Credentials import credential_store
from Ganga.Core.exceptions import InvalidCredentialError

_allShells = {}

logger = getLogger()

def constructShell():
    """
    Construct a grid shell based upon either the GLITE_SETUP or GLITE_LOCATION as possibly defined by the user
    """

    values = {}
    for key in ['X509_CERT_DIR', 'X509_VOMS_DIR']:
        try:
            values[key] = os.environ[key]
        except KeyError:
            pass

    config = getConfig('LCG')

    # 1. check if the GLITE_SETUP is changed by user -> take the user's value as session value
    # 2. else check if GLITE_LOCATION is defined as env. variable -> do nothing (ie. create shell without any lcg setup)
    # 3. else take the default GLITE_SETUP as session value

    if config.getEffectiveLevel('GLITE_SETUP') == 2 and 'GLITE_LOCATION' in os.environ:
        s = Shell()
    else:
        if os.path.exists(config['GLITE_SETUP']):
            s = Shell(config['GLITE_SETUP'])
        else:
            logger.error("Configuration of GLITE for LCG: ")
            logger.error("File not found: %s" % config['GLITE_SETUP'])
            return None

    for key, val in values.items():
        s.env[key] = val

    # check and set env. variables for default LFC setup
    if 'LFC_HOST' not in s.env:
        try:
            s.env['LFC_HOST'] = config['DefaultLFC']
        except ConfigError:
            pass

    if 'LFC_CONNTIMEOUT' not in s.env:
        s.env['LFC_CONNTIMEOUT'] = '20'

    if 'LFC_CONRETRY' not in s.env:
        s.env['LFC_CONRETRY'] = '0'

    if 'LFC_CONRETRYINT' not in s.env:
        s.env['LFC_CONRETRYINT'] = '1'

    return s

def getShell(cred_req=None):
    """
    Utility function for getting Grid Shell.

    If a cred_req is given then the grid shell which has been cached for this credential requirement is returned.
    If a cred_req is given an the credential doesn't exist in the credential_store then an InvalidCredentialError exception is raised

    If no cred_req is given then a grid shell is contructed based upon either the GLITE_SETUP or GLITE_LOCATION as possibly defined by the user
        THERE IS NO CACHING MADE HERE IN THIS CASE!!!

    Arguments:
        cred_req (ICredentialRequirement): This is the credential requirement required.
    """

    if cred_req is not None:
        if not credential_store[cred_req].is_valid():
            logger.info('GridShell.getShell given credential which is invalid')
            raise InvalidCredentialError()

        if cred_req in _allShells.keys():
            return _allShells[cred_req]

    constructed_shell = constructShell()

    if cred_req is not None:
        constructed_shell.env['X509_USER_PROXY'] = credential_store[cred_req].location

    _allShells[cred_req] = constructed_shell

    return constructed_shell



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ClassifyLCG.py,v 1.1 2008-10-04 17:42:38 karl Exp $
###############################################################################
# File: ClassifyLCG.py
# Author: K. Harrison
# Created: 070126

"""Module containing class dealing with preparation of jobs to run
   Classify application on LCG backend"""

__author__  = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__    = "30 November 2007"
__version__ = "1.1"

from ClassifyLocal import ClassifyLocal
from ClassifyLocal import ptime
from Ganga.GPIDev.Lib.File import  File
from Ganga.Lib.LCG import LCGJobConfig
from Ganga.Lib.LCG import LCGRequirements
from Ganga.Utility import logging
from Ganga.Utility.files import fullpath

import os

logger = logging.getLogger()

class ClassifyLCG( ClassifyLocal ):

   def prepare( self, app, appsubconfig, appmasterconfig, jobmasterconfig ):

      local = ClassifyLocal.prepare\
         ( self, app, appsubconfig, appmasterconfig, jobmasterconfig )
      inbox = []
      inbox.extend( local.inputbox )
      requirements = LCGRequirements()
      tag = "VO-camont-%s" % appsubconfig[ "version" ]
      requirements.software = [ tag ]
      lcg = LCGJobConfig( exe = local.exe, inputbox = inbox,\
         outputbox = local.outputbox, requirements = requirements )
      return lcg

   def tail( self, job ):
      lineList, outbox = ClassifyLocal.tail( self )
     # Remove last lines (completion times) for local jobs
      lineList = lineList[ : -4 ]
     # Remove tarball from outbox, as this is to be uploaded to storage element
      outbox.remove( "images.tar.gz" ) 

      gridUrl = os.path.join( job.outputdata.getGridStorage(), "images.tar.gz" )
      lineList.extend\
         ( [
         "",
         "UPLOAD_START_TIME=\\",
         "%s" % ptime,
         "edg-gridftp-mkdir --parents %s" % os.path.dirname( gridUrl ),
         "globus-url-copy file:${WORKDIR}/images.tar.gz %s" % gridUrl,
         "UPLOAD_END_TIME=\\",
         "%s" % ptime,
         "",
         "echo \"Upload_start: ${UPLOAD_START_TIME}\" >> ${RUN_DATA}",
         "echo \"Upload_end: ${UPLOAD_END_TIME}\" >> ${RUN_DATA}",
         "JOB_END_TIME=\\",
         "%s" % ptime,
         "echo \"Job_end: ${JOB_END_TIME}\" >> ${RUN_DATA}",
         "echo \"End time: $(date)\"",
         ] )
      return ( lineList, outbox )



"""
Provides Persistency and the base class for all Registries
Also, a list of all Registries is kept here
"""

from Ganga.Core.GangaRepository.GangaRepository import GangaRepository
from Ganga.Core.exceptions import RepositoryError, InaccessibleObjectError, SchemaVersionError
from Ganga.Core.GangaRepository.Registry import RegistryError, RegistryAccessError, RegistryKeyError, RegistryLockError, ObjectNotInRegistryError
from Ganga.Core.GangaRepository import GangaRepositoryXML

allRegistries = {}


def addRegistry(registry):
    """ Add a registry to the global dict"""
    allRegistries[registry.name] = registry

def getRegistries():
    """ Get all registries from the global dict"""
    return allRegistries.values()

def getRegistry(name):
    """ Get a specific Registry from the global dict"""
    return allRegistries[name]

def getRegistrySlice(name):
    """ Get The registry slice wrapping the registry of choice"""
    return allRegistries[name].getSlice()

def getRegistryProxy(name):
    """ Get the proxied registry slice wrapping the registry of choice"""
    return allRegistries[name].getProxy()




##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: FileBuffer.py,v 1.1 2008-07-17 16:40:53 moscicki Exp $
##########################################################################

import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()

import os


class FileBuffer(object):

    """ FileBuffer represents a file in memory which has not been yet created.
        This is a handy way of creating small wrapper scripts to be generated on the fly.
    """

    # Added a subdir (see File.py for comments) - AM
    def __init__(self, name, contents, subdir=os.curdir, executable=0):
        """ name is the name of the file to be created
            contents is the text with file contents or a file-object which will be read()
            executable indicates if a file is create()'ed with executable permissions
        """
        self.name = name
        self._contents = contents
        self.subdir = subdir
        self.executable = executable

    def getPathInSandbox(self):
        """return a relative location of a file in a sandbox: subdir/name"""
        from Ganga.Utility.files import real_basename
        return os.path.join(self.subdir, real_basename(self.name))

    def getContents(self):
        """return a string with the contents of the file buffer"""
        logger.debug("Reading FileBuffer: %s" % self.name)
        if isinstance(self._contents, str):
            return self._contents
        else:
            return self._contents.read()

    def create(self, outname=None):
        """create a file in a local filesystem as 'outname' """
        filename = self.name
        if outname is not None:
            filename = outname

        with open(filename, 'w') as thisfile:
            thisfile.write(self.getContents())

        if self.executable:
            from Ganga.Utility.files import chmod_executable
            chmod_executable(filename)

        logger.debug("Created %s in: %s" %
                     (filename, os.path.realpath(filename)))

        return self

    def isExecutable(self):
        """ return true if a file is create()'ed with executable permissions"""
        return self.executable


#
#
#
# $Log: not supported by cvs2svn $
# Revision 1.5  2007/08/28 08:26:35  moscicki
# fixed typo
#
# Revision 1.4  2005/08/23 17:07:23  moscicki
# Added executable flag for FileBuffer.
# Added create method for File and FileBuffer.
# Added getPathInSandbox() method.
#
# Revision 1.3  2005/08/10 09:45:36  andrew
# Added a subdir to File and FileBuffer objects. Changed the writefile method
# in FileWorspace to use the subdirectory
#
#
#
#




from Ganga.Utility.Config import getConfig

outputconfig = getConfig("Output")

def getSharedPath():
    # Required to be initialized for ShareDir object
    from Ganga.Utility.files import expandfilename
    import os.path
    Conf_config = getConfig('Configuration')
    root_default = os.path.join(expandfilename(Conf_config['gangadir']), 'shared', Conf_config['user'])
    return root_default





from Ganga.Core.exceptions import BackendError, ApplicationConfigurationError
from Ganga.Core.exceptions import SplitterError
from Ganga.Utility.logging import getLogger
from Ganga.Utility.Config import getConfig
from Ganga.Utility.util import unique
from GangaDirac.Lib.Splitters.SplitterUtils import DiracSplitter
from GangaDirac.Lib.Files.DiracFile import DiracFile
from Ganga.GPIDev.Base.Proxy import getName
logger = getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def mangle_job_name(app):
    """ Create a safe job name to send to DIRAC (includes full fqid)
    Args:
        app (IApplication): This is the application belonging to the job of interest
    """
    job = app.getJobObject()

    jobName = job.name
    jobIndex = job.getStringFQID()
    appName = getName(app)
    appVersion = None
    if hasattr(app, 'version'):
        appVersion = str(app.version)

    result = ''
    addBracket = False
    if jobName:
        result += '%s__' % jobName
    if appName:
        # and not j.subjobs: #not necessary?
        if not job.master and job.splitter:
            result += '{Ganga_%s_(%s.%s)' % (appName, jobIndex, '%n')
        else:
            result += '{Ganga_%s_(%s)' % (appName, jobIndex)
        addBracket = True
    if appVersion:
        result += '_%s' % appVersion
    if not result:
        result = '{Ganga_Job}'
    elif addBracket:
        result += '}'
    return result
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def API_nullifier(item):
    """ Return is item None or emtpy list
    Args:
        item (list): If this is an empty list return None, else return list, Dirac doesn't like empty lists
    """
    if item is None or len(item) == 0:
        return None
    return item

def dirac_outputfile_jdl(output_files, empty_SE_check):
    """
    This constructs the setOutputData such that the data will be sent to the chosen SE/Token
    In the case that the empty_SE_check is True it will raise an exception if the defaultSE is empty
    In the case that it's False an empty SE is allowed.
    Args:
        output_files (list): List of IGangaFile objects which are requested from job.outputfiles
        empty_SE_check (bool): If this is True then throw exception if DiracFile objects don't have any defaultSE set
    """

    _output_files = [this_file for this_file in output_files if isinstance(this_file, DiracFile)]

    file_SE_dict = {}

    for this_file in _output_files:
        if not this_file.defaultSE in file_SE_dict:
            file_SE_dict[this_file.defaultSE] = []
        file_SE_dict[this_file.defaultSE].append( this_file.namePattern )

    per_SE_JDL = '''j.setOutputData(###OUTPUTDATA###, outputPath='###OUTPUT_PATH###', outputSE=###OUTPUT_SE###)'''
    total_JDL = ''

    for outputSE, namePatterns in file_SE_dict.iteritems():

        myLine = str(per_SE_JDL)
        myLine = myLine.replace('###OUTPUTDATA###', str(namePatterns))
        if outputSE != '':
            myLine = myLine.replace('###OUTPUT_SE###', str([outputSE]))
        else:
            if empty_SE_check:
                ## If true check, if not false check
                raise BackendError("Dirac", "Can't submit a DIRAC job with DiracFile outputfile without setting a defaultSE.")
            myLine = myLine.replace('###OUTPUT_SE###', str([]))

        total_JDL += myLine + "\n"

    return total_JDL


def dirac_inputdata(app, hasOtherInputData=False):
    """ Construct the JDL component which requests the inputdata for a job
    Args:
        app (IApplication): app which belongs to the job of interest
        hasOtherInputData (bool): This is used to stop BannedSites being added to the JDL structure through backend.settings
    """
    job = app.getJobObject()
    input_data = None
    parametricinput_data = None

    inputLFNs = []

    input_data = None
    parametricinput_data = None

    if not job.inputdata and (not job.master or not job.master.inputdata):
        return input_data, parametricinput_data

    wanted_job = job
    if not job.inputdata and job.master and job.master.inputdata is not None and job.master.inputdata:
        wanted_job = job.master

    inputLFNs = ['LFN:'+this_file.lfn for this_file in wanted_job.inputdata if isinstance(this_file, DiracFile)]

    # master job with a splitter reaching prepare, hence bulk submit
    if not job.master and job.splitter:
        parametricinput_data = dirac_parametric_split(app)
        if parametricinput_data is not None and len(parametricinput_data) > getConfig('DIRAC')['MaxDiracBulkJobs']:
            raise BackendError('Dirac', 'Number of bulk submission jobs \'%s\' exceeds the maximum allowed \'%s\' if more are needed please modify your config. Note there is a hard limit in Dirac of currently 1000.' % (
                len(parametricinput_data), getConfig('DIRAC')['MaxDiracBulkJobs']))
        # master job with no splitter or subjob already split proceed as normal
        else:
            input_data = inputLFNs

    if 'Destination' not in job.backend.settings and not inputLFNs and not hasOtherInputData:
        t1_sites = getConfig('DIRAC')['noInputDataBannedSites']
        logger.info('Job has no inputdata (T1 sites will be banned to help avoid overloading them).')
        if 'BannedSites' in job.backend.settings:
            job.backend.settings['BannedSites'].extend(t1_sites)
            job.backend.settings['BannedSites'] = unique(job.backend.settings['BannedSites'])
        else:
            if t1_sites:
                job.backend.settings['BannedSites'] = t1_sites[:]

    if not input_data and not parametricinput_data:
        input_data = inputLFNs

    return input_data, parametricinput_data

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def dirac_parametric_split(app):
    """ Bulk job submission splitter. TODO document more
    Args:
        app (IApplication): Application belonging to the job in question
    """
    data = app.getJobObject().inputdata
    splitter = app.getJobObject().splitter

    split_data = [dataset for dataset in DiracSplitter(data, splitter.filesPerJob, splitter.maxFiles, splitter.ignoremissing)]

    split_files = []

    for dataset in split_data:
        this_dataset = []
        for this_file in dataset:
            if isinstance(this_file, DiracFile):
                this_dataset.append(this_file.lfn)
            else:
                raise SplitterError("ERROR: file: %s NOT of type DiracFile" % str(this_file) )
        split_files.append(this_dataset)

    if len(split_files) > 0:
        return split_files

    return None


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
def dirac_ouputdata(app):
    """ TODO work out if this is still called anywhere?
    Returns the outputdata files as a tuple of files and location
    Args:
        app (IApplication): App for the job of interest
    """
    job = app.getJobObject()
    if job.outputdata:
        return job.outputdata.files[:], job.outputdata.location
    return None, None


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
def diracAPI_script_template():
    """ Generate and return the DiracAPI job submission template """

    import inspect
    import os.path
    script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))),
                                   'DiracRTHScript.py.template')

    from Ganga.GPIDev.Lib.File import FileUtils
    script_template = FileUtils.loadScript(script_location, '')

    return script_template

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def diracAPI_script_settings(app):
    """
    Set some additional setting on the diracAPI in the JDL making use of any custom parameters set in the backend object
    return JDL lines
    Args:
        app (IApplication): Application belonging to job of interest
    """
    job = app.getJobObject()
    diracAPI_line = ''
    if type(job.backend.settings) is not dict:
        raise ApplicationConfigurationError(
            None, 'backend.settings should be a dict')
    for setting, setting_val in job.backend.settings.iteritems():
        if str(setting).startswith('set'):
            _setting = str(setting)[3:]
        else:
            _setting = str(setting)
        if type(setting_val) is str:
            setting_line = 'j.set###SETTING###("###VALUE###")\n'
        else:
            setting_line = 'j.set###SETTING###(###VALUE###)\n'
        diracAPI_line += setting_line.replace('###SETTING###', _setting).replace('###VALUE###', str(setting_val))
    if diracAPI_line == '':
        diracAPI_line = None

    return diracAPI_line
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#




from Ganga.Utility.logging import getLogger
from Ganga.Core.exceptions import GangaValueError
logger = getLogger()


class PluginManagerError(GangaValueError):

    def __init__(self, x):
        super(PluginManagerError, self).__init__(self, x)

# Simple Ganga Plugin Mechanism
#
# Any object may be registered (added) in the plugin manager provided that
# you are able to specify the name and the category to which it belongs.
#
# If you do not use category all plugins are registered in a flat list. Otherwise
# there is a list of names for each category seaprately.


class PluginManager(object):

    def __init__(self):
        self.all_dict = {}
        self.first = {}
        self._prev_found = {}

    def find(self, category, name):
        """
        Return a plugin added with 'name' in the given 'category'.
        If 'name' is None then the default plugin in the category is returned.
        Typically the default plugin is the first added.
        If plugin not found raise PluginManagerError.
        """
        #logger.debug( "Attempting to Find Plugin: %s" % name )
        #import traceback
        # traceback.print_stack()

        # Simple attempt to pre-load and cache Plugin lookups
        key = str(category) + "_" + str(name)
        if key in self._prev_found:
            return self._prev_found[key]

        try:
            if name is not None:
                if category in self.first:
                    ## This is expected to work and is quite verbose when debugging turned on
                    #logger.debug("Returning based upon Category and Name")
                    #logger.debug("name: %s cat: %s" % (str(name), str(category)))
                    if name in self.all_dict[category]:
                        self._prev_found[key] = self.all_dict[category][name]
                        return self.all_dict[category][name]

            if (name is None) and category is not None:
                if (category in self.first):
                    ## This is expected to work and is quite verbose when debugging turned on
                    #logger.debug("Returning based upon Category ONLY")
                    #logger.debug("name: %s cat: %s" % (str(name), str(category)))
                    self._prev_found[key] = self.first[category]
                    return self.first[category]

            elif (name is not None) and (category is not None):
                for category_i in self.all_dict:
                    for this_name in self.all_dict[category_i]:
                        if name == this_name:
                            message1 = "Category of %s, has likely changed between ganga versions!" % name
                            message2 = "Category Requested: %s,   Category in which plugin was found: %s" % (category, category_i)
                            message3 = "Attempting to use new category %s to load a stored object, this may fail!" % category_i
                            logger.debug(message1)
                            logger.debug(message2)
                            logger.debug(message3)
                            self._prev_found[key] = self.all_dict[category_i][name]
                            return self.all_dict[category_i][name]

        except KeyError:
            logger.debug("KeyError from Config system!")
        except:
            logger.error("Some Other unexpected ERROR!")
            raise

        if name is None:
            s = "cannot find default plugin for category " + category
        else:
            s = "cannot find '%s' in a category '%s', or elsewhere" % (name, category)

        if name is None and category is None:
            s = "Serious Plugin Error has occured"

        logger.debug(s)
        raise PluginManagerError(s)

    def add(self, pluginobj, category, name):
        """ Add a pluginobj to the plugin manager with the name and the category labels.
        The first plugin is default unless changed explicitly.
        """
        cat = self.all_dict.setdefault(category, {})
        self.first.setdefault(category, pluginobj)
        cat[name] = pluginobj
        logger.debug('adding plugin %s (category "%s") ' % (name, category))

    def setDefault(self, category, name):
        """ Make the plugin 'name' be default in a given 'category'.
        You must first add() the plugin object before calling this method. Otherwise
        PluginManagerError is raised.
        """
        assert(not name is None)
        pluginobj = self.find(category, name)
        self.first[category] = pluginobj

    def allCategories(self):
        return self.all_dict

    def allClasses(self, category):
        cat = self.all_dict.get(category)
        if cat:
            return cat
        else:
            return {}

allPlugins = PluginManager()




###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: Condor.py,v 1.8 2009/04/02 17:52:24 karl Exp $
###############################################################################
# File: Condor.py
# Author: K. Harrison
# Created: 051228
#
# KH - 060321: Additions from Johannes Elmsheuser, to allow use with Condor-G
#
# KH - 060728: Changes for framework migration
#
# KH - 061012: Updates in submit and preparejob methods, for core changes
#
# KH - 061027: Corrections for case of no shared filesystem
#
# KH - 080213: Changes made to use global Condor id for tracking job status
#
# KH - 080215: Correction for passing quoted strings as a command-line argument
#              to an executable
#
# KH - 080410: Corrections made to kill() method of Condor class
#
# KH - 080410: Implemented resubmit() method of Condor class
#
# KH - 080729: Updates for changes to JobConfig class in Ganga 5
#              Error message printed in case submit command fails
#
# KH - 081008 : Added typelist information for schema property "submit_options"
#
# KH - 081102 : Remove spurious print statement
#
# KH - 090128 : Added getenv property to Condor class, to allow environment of
#               submit machine to be passed to worker node
#
#               Added creation of bash job wrapper, to allow environment
#               setup by defining BASH_ENV to point to setup script
#
#               Set status to failed for job with non-zero exit code
#
# KH - 090307 : Modified kill() method to assume success even in case of
#               non-zero return code from condor_rm
#
# KH - 090308 : Modified updateMonitoringInformation() method
#               to deal with case where all queues are empty
#
# KH - 090402 : Corrected bug that meant application arguments were ignored
#
#               In script to be run on worker node, print warning if unable
#               to find startup script pointed to by BASH_ENV
#
# KH - 090809 : Changed logic for updating job status to final value -
#               Condor log file is now searched to check whether job
#               is marked as "aborted" or "terminated"

"""Module containing class for handling job submission to Condor backend"""

__author__ = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__ = "09 August 2009"
__version__ = "2.5"

from Ganga.Core import Sandbox
from Ganga.GPIDev.Adapters.IBackend import IBackend
from Ganga.GPIDev.Base.Proxy import stripProxy
from Ganga.GPIDev.Lib.File.FileBuffer import FileBuffer
from Ganga.GPIDev.Schema import ComponentItem, Schema, SimpleItem, Version
from Ganga.Utility.ColourText import Foreground, Effects

import Ganga.Utility.logging
from Ganga.Utility.Config import getConfig

import commands
import inspect
import os
import shutil
import time

logger = Ganga.Utility.logging.getLogger()


class Condor(IBackend):

    """Condor backend - submit jobs to a Condor pool.

    For more options see help on CondorRequirements.
    """

    _schema = Schema(Version(1, 0), {
        "requirements": ComponentItem(category="condor_requirements",
                                      defvalue="CondorRequirements",
                                      doc="Requirements for selecting execution host"),
        "env": SimpleItem(defvalue={},
                          doc='Environment settings for execution host'),
        "getenv": SimpleItem(defvalue="False",
                             doc='Flag to pass current envrionment to execution host'),
        "rank": SimpleItem(defvalue="Memory",
                           doc="Ranking scheme to be used when selecting execution host"),
        "submit_options": SimpleItem(defvalue=[], typelist=[str],
                                     sequence=1, doc="Options passed to Condor at submission time"),
        "id": SimpleItem(defvalue="", protected=1, copyable=0,
                         doc="Condor jobid"),
        "status": SimpleItem(defvalue="", protected=1, copyable=0,
                             doc="Condor status"),
        "cputime": SimpleItem(defvalue="", protected=1, copyable=0,
                              doc="CPU time used by job"),
        "actualCE": SimpleItem(defvalue="", protected=1, copyable=0,
                               doc="Machine where job has been submitted"),
        "shared_filesystem": SimpleItem(defvalue=True,
                                        doc="Flag indicating if Condor nodes have shared filesystem"),
        "universe": SimpleItem(defvalue="vanilla",
                               doc="Type of execution environment to be used by Condor"),
        "globusscheduler": SimpleItem(defvalue="", doc="Globus scheduler to be used (required for Condor-G submission)"),
        "globus_rsl": SimpleItem(defvalue="",
                                 doc="Globus RSL settings (for Condor-G submission)"),
        "accounting_group": SimpleItem(defvalue='', doc="Provide an accounting group for this job."),
        "cdf_options": SimpleItem(defvalue={}, doc="Additional options to set in the CDF file given by a dictionary")
    })

    _category = "backends"
    _name = "Condor"
    statusDict = \
        {
            "0": "Unexpanded",
            "1": "Idle",
            "2": "Running",
            "3": "Removed",
            "4": "Completed",
            "5": "Held"
        }

    def __init__(self):

        # Add a volatile variable for recording the first time a job's stdout is checked
        self._stdout_check_time = 0

        super(Condor, self).__init__()

    def submit(self, jobconfig, master_input_sandbox):
        """Submit job to backend.

            Return value: True if job is submitted successfully,
                          or False otherwise"""

        cdfpath = self.preparejob(jobconfig, master_input_sandbox)
        status = self.submit_cdf(cdfpath)
        return status

    def submit_cdf(self, cdfpath=""):
        """Submit Condor Description File.

            Argument other than self:
               cdfpath - path to Condor Description File to be submitted

            Return value: True if job is submitted successfully,
                          or False otherwise"""

        commandList = ["condor_submit -v"]
        commandList.extend(self.submit_options)
        commandList.append(cdfpath)
        commandString = " ".join(commandList)

        status, output = commands.getstatusoutput(commandString)

        self.id = ""
        if 0 != status:
            logger.error\
                ("Tried submitting job with command: '%s'" % commandString)
            logger.error("Return code: %s" % str(status))
            logger.error("Condor output:")
            logger.error(output)
        else:
            tmpList = output.split("\n")
            for item in tmpList:
                if 1 + item.find("** Proc"):
                    localId = item.strip(":").split()[2]
                    queryCommand = " ".join\
                        (["condor_q -format \"%s\" GlobalJobId", localId])
                    qstatus, qoutput = commands.getstatusoutput(queryCommand)
                    if 0 != status:
                        logger.warning\
                            ("Problem determining global id for Condor job '%s'" %
                             localId)
                        self.id = localId
                    else:
                        self.id = qoutput
                    break

        return not self.id is ""

    def resubmit(self):
        """Resubmit job that has already been configured.

            Return value: True if job is resubmitted successfully,
                          or False otherwise"""

        job = self.getJobObject()

        inpDir = job.getInputWorkspace().getPath()
        outDir = job.getOutputWorkspace().getPath()

        # Delete any existing output files, and recreate output directory
        if os.path.isdir(outDir):
            shutil.rmtree(outDir)
        if os.path.exists(outDir):
            os.remove(outDir)
        os.mkdir(outDir)

        # Determine path to job's Condor Description File
        cdfpath = os.path.join(inpDir, "__cdf__")

        # Resubmit job
        if os.path.exists(cdfpath):
            status = self.submit_cdf(cdfpath)
        else:
            logger.warning\
                ("No Condor Description File for job '%s' found in '%s'" %
                 (str(job.id), inpDir))
            logger.warning("Resubmission failed")
            status = False

        return status

    def kill(self):
        """Kill running job.

           No arguments other than self

           Return value: True if job killed successfully,
                         or False otherwise"""

        job = self.getJobObject()

        if not self.id:
            logger.warning("Job %s not running" % job.id)
            return False

        idElementList = job.backend.id.split("#")
        if 3 == len(idElementList):
            if idElementList[1].find(".") != -1:
                killCommand = "condor_rm -name %s %s" % \
                    (idElementList[0], idElementList[1])
            else:
                killCommand = "condor_rm -name %s %s" % \
                    (idElementList[0], idElementList[2])
        else:
            killCommand = "condor_rm %s" % (idElementList[0])

        status, output = commands.getstatusoutput(killCommand)

        if (status != 0):
            logger.warning\
                ("Return code '%s' killing job '%s' - Condor id '%s'" %
                 (str(status), job.id, job.backend.id))
            logger.warning("Tried command: '%s'" % killCommand)
            logger.warning("Command output: '%s'" % output)
            logger.warning("Anyway continuing with job removal")

        job.backend.status = "Removed"
        killStatus = True

        return killStatus

    def preparejob(self, jobconfig, master_input_sandbox):
        """Prepare Condor description file"""

        job = self.getJobObject()
        inbox = job.createPackedInputSandbox(jobconfig.getSandboxFiles())
        inpDir = job.getInputWorkspace().getPath()
        outDir = job.getOutputWorkspace().getPath()

        infileList = []

        exeString = jobconfig.getExeString().strip()
        quotedArgList = []
        for arg in jobconfig.getArgStrings():
            quotedArgList.append("\\'%s\\'" % arg)
        exeCmdString = " ".join([exeString] + quotedArgList)

        for filePath in inbox:
            if not filePath in infileList:
                infileList.append(filePath)

        for filePath in master_input_sandbox:
            if not filePath in infileList:
                infileList.append(filePath)

        fileList = []
        for filePath in infileList:
            fileList.append(os.path.basename(filePath))

        if job.name:
            name = job.name
        else:
            name = job.application._name
        name = "_".join(name.split())
        wrapperName = "_".join(["Ganga", str(job.id), name])

        commandList = [
            "#!/usr/bin/env python",
            "from __future__ import print_function",
            "# Condor job wrapper created by Ganga",
            "# %s" % (time.strftime("%c")),
            "",
            inspect.getsource(Sandbox.WNSandbox),
            "",
            "import os",
            "import time",
            "import mimetypes",
            "import shutil",
            "",
            "startTime = time.strftime"
            + "( '%a %d %b %H:%M:%S %Y', time.gmtime( time.time() ) )",
            "",
            "for inFile in %s:" % str(fileList),
            "   if mimetypes.guess_type(inFile)[1] in ['gzip', 'bzip2']:",
            "       getPackedInputSandbox( inFile )",
            "   else:",
            "       shutil.copy(inFile, os.path.join(os.getcwd(), os.path.basename(inFile)))",
            "",
            "exePath = '%s'" % exeString,
            "if os.path.isfile( '%s' ):" % os.path.basename(exeString),
            "   os.chmod( '%s', 0755 )" % os.path.basename(exeString),
            "wrapperName = '%s_bash_wrapper.sh'" % wrapperName,
            "wrapperFile = open( wrapperName, 'w' )",
            "wrapperFile.write( '#!/bin/bash\\n' )",
            "wrapperFile.write( 'echo \"\"\\n' )",
            "wrapperFile.write( 'echo \"Hostname: $(hostname -f)\"\\n' )",
            "wrapperFile.write( 'echo \"\\${BASH_ENV}: ${BASH_ENV}\"\\n' )",
            "wrapperFile.write( 'if ! [ -z \"${BASH_ENV}\" ]; then\\n' )",
            "wrapperFile.write( '  if ! [ -f \"${BASH_ENV}\" ]; then\\n' )",
            "wrapperFile.write( '    echo \"*** Warning: "
            + "\\${BASH_ENV} file not found ***\"\\n' )",
            "wrapperFile.write( '  fi\\n' )",
            "wrapperFile.write( 'fi\\n' )",
            "wrapperFile.write( 'echo \"\"\\n' )",
            "wrapperFile.write( '%s\\n' )" % exeCmdString,
            "wrapperFile.write( 'exit ${?}\\n' )",
            "wrapperFile.close()",
            "os.chmod( wrapperName, 0755 )",
            "result = os.system( './%s' % wrapperName )",
            "os.remove( wrapperName )",
            "",
            "endTime = time.strftime"
              + "( '%a %d %b %H:%M:%S %Y', time.gmtime( time.time() ) )",
            "print('\\nJob start: ' + startTime)",
            "print('Job end: ' + endTime)",
            "print('Exit code: %s' % str( result ))"
        ]

        commandString = "\n".join(commandList)
        wrapper = job.getInputWorkspace().writefile\
            (FileBuffer(wrapperName, commandString), executable=1)

        infileString = ",".join(infileList)
        outfileString = ",".join(jobconfig.outputbox)

        cdfDict = \
            {
                'universe': self.universe,
                'on_exit_remove': 'True',
                'should_transfer_files': 'YES',
                'when_to_transfer_output': 'ON_EXIT_OR_EVICT',
                'executable': wrapper,
                'transfer_executable': 'True',
                'notification': 'Never',
                'rank': self.rank,
                'initialdir': outDir,
                'error': 'stderr',
                'output': 'stdout',
                'log': 'condorLog',
                'stream_output': 'false',
                'stream_error': 'false',
                'getenv': self.getenv
            }

        # extend with additional cdf options
        cdfDict.update(self.cdf_options)

        # accounting group
        if self.accounting_group:
            cdfDict['accounting_group'] = self.accounting_group

        envList = []
        if self.env:
            for key in self.env.keys():
                value = self.env[key]
                if (isinstance(value, str)):
                    value = os.path.expandvars(value)
                else:
                    value = str(value)
                envList.append("=".join([key, value]))
        envString = ";".join(envList)
        if jobconfig.env:
            for key in jobconfig.env.keys():
                value = jobconfig.env[key]
                if (isinstance(value, str)):
                    value = os.path.expandvars(value)
                else:
                    value = str(value)
                envList.append("=".join([key, value]))
        envString = ";".join(envList)
        if envString:
            cdfDict['environment'] = envString

        if infileString:
            cdfDict['transfer_input_files'] = infileString

        if self.globusscheduler:
            cdfDict['globusscheduler'] = self.globusscheduler

        if self.globus_rsl:
            cdfDict['globus_rsl'] = self.globus_rsl

        if outfileString:
            cdfDict['transfer_output_files'] = outfileString

        cdfList = [
            "# Condor Description File created by Ganga",
            "# %s" % (time.strftime("%c")),
            ""]
        for key, value in cdfDict.iteritems():
            cdfList.append("%s = %s" % (key, value))
        cdfList.append(self.requirements.convert())
        cdfList.append("queue")
        cdfString = "\n".join(cdfList)

        return job.getInputWorkspace().writefile\
            (FileBuffer("__cdf__", cdfString))

    def updateMonitoringInformation(jobs):

        jobDict = {}
        for job in jobs:
            if job.backend.id:
                jobDict[job.backend.id] = job

        idList = jobDict.keys()

        if not idList:
            return

        queryCommand = " ".join\
            ([
                "condor_q -global" if getConfig(
                    "Condor")["query_global_queues"] else "condor_q",
                "-format \"%s \" GlobalJobId",
                "-format \"%s \" RemoteHost",
                "-format \"%d \" JobStatus",
                "-format \"%f\\n\" RemoteUserCpu"
            ])
        status, output = commands.getstatusoutput(queryCommand)
        if 0 != status:
            logger.error("Problem retrieving status for Condor jobs")
            return

        if ("All queues are empty" == output):
            infoList = []
        else:
            infoList = output.split("\n")

        allDict = {}
        for infoString in infoList:
            tmpList = infoString.split()
            id, host, status, cputime = ("", "", "", "")
            if 3 == len(tmpList):
                id, status, cputime = tmpList
            if 4 == len(tmpList):
                id, host, status, cputime = tmpList
            if id:
                allDict[id] = {}
                allDict[id]["status"] = Condor.statusDict[status]
                allDict[id]["cputime"] = cputime
                allDict[id]["host"] = host

        fg = Foreground()
        fx = Effects()
        status_colours = {'submitted': fg.orange,
                          'running': fg.green,
                          'completed': fg.blue}

        for id in idList:

            printStatus = False
            if jobDict[id].status == "killed":
                continue

            localId = id.split("#")[-1]
            globalId = id

            if globalId == localId:
                queryCommand = " ".join\
                    ([
                        "condor_q -global" if getConfig(
                            "Condor")["query_global_queues"] else "condor_q",
                        "-format \"%s\" GlobalJobId",
                        id
                    ])
                status, output = commands.getstatusoutput(queryCommand)
                if 0 == status:
                    globalId = output

            if globalId in allDict.keys():
                status = allDict[globalId]["status"]
                host = allDict[globalId]["host"]
                cputime = allDict[globalId]["cputime"]
                if status != jobDict[id].backend.status:
                    printStatus = True
                    stripProxy(jobDict[id])._getSessionLock()
                    jobDict[id].backend.status = status
                    if jobDict[id].backend.status == "Running":
                        jobDict[id].updateStatus("running")

                if host:
                    if jobDict[id].backend.actualCE != host:
                        jobDict[id].backend.actualCE = host
                jobDict[id].backend.cputime = cputime
            else:
                jobDict[id].backend.status = ""
                outDir = jobDict[id].getOutputWorkspace().getPath()
                condorLogPath = "".join([outDir, "condorLog"])
                checkExit = True
                if os.path.isfile(condorLogPath):
                    checkExit = False
                    for line in open(condorLogPath):
                        if -1 != line.find("terminated"):
                            checkExit = True
                            break
                        if -1 != line.find("aborted"):
                            checkExit = True
                            break

                if checkExit:
                    printStatus = True
                    stdoutPath = "".join([outDir, "stdout"])
                    jobStatus = "failed"
                    if os.path.isfile(stdoutPath):
                        with open(stdoutPath) as stdout:
                            lineList = stdout.readlines()
                        try:
                            exitLine = lineList[-1]
                            exitCode = exitLine.strip().split()[-1]
                        except IndexError:
                            exitCode = '-1'

                        if exitCode.isdigit():
                            jobStatus = "completed"
                        else:
                            # Some filesystems/setups have the file created but empty - only worry if it's been 10mins
                            # since we first checked the file
                            if len(lineList) == 0:
                                if not jobDict[id].backend._stdout_check_time:
                                    jobDict[id].backend._stdout_check_time = time.time()

                                if (time.time() - jobDict[id].backend._stdout_check_time) < 10*60:
                                    continue
                                else:
                                    logger.error("Empty stdout file from job %s after waiting 10mins. Marking job as"
                                                 "failed." % jobDict[id].fqid)
                            else:
                                logger.error("Problem extracting exit code from job %s. Line found was '%s'." % (
                                    jobDict[id].fqid, exitLine))

                    jobDict[id].updateStatus(jobStatus)

            if printStatus:
                if jobDict[id].backend.actualCE:
                    hostInfo = jobDict[id].backend.actualCE
                else:
                    hostInfo = "Condor"
                status = jobDict[id].status
                if status in status_colours:
                    colour = status_colours[status]
                else:
                    colour = fg.magenta
                if "submitted" == status:
                    preposition = "to"
                else:
                    preposition = "on"

                if jobDict[id].backend.status:
                    backendStatus = "".join\
                        ([" (", jobDict[id].backend.status, ") "])
                else:
                    backendStatus = ""

                logger.info(colour + 'Job %s %s%s %s %s - %s' + fx.normal,
                            jobDict[
                                id].fqid, status, backendStatus, preposition, hostInfo,
                            time.strftime('%c'))

        return None

    updateMonitoringInformation = \
        staticmethod(updateMonitoringInformation)

#_________________________________________________________________________


